/* Tehnica RAII
   Resource Aquisition Is Initialization

   === The Rule of Three ===

   Daca suntem obligati sa implementam una dintre
   urmatoarele 3 functii  membre speciale:
    1. Constructor + Constructor de Copiere
    2. Destructor
    3. Operator de Atribuire

    atunci SUNTEM OBLIGATI sa le IMPLEMENTAM PE TOATE
*/
#include <iostream>

template <class T>
class Vector {
public:

 // Tipuri membre
    using iterator = T*;

    Vector() : n {0}, ptr {nullptr}
    {
    }
//========== RAII================================/
    // In contructor se achizitioneaza resursele
    Vector(size_t count, const T& val = T())
        : n {count}, ptr {new T[n]}
    {
        for (size_t i = 0; i < n; ++i)
            ptr[i] = val;
       
    }

//	Vector(const Vector&) = delete; // interzice copiere
    // Copy-Constructor
    Vector(const Vector& v) :
        n {v.n}, ptr {new T[n]}
    {
        for (size_t i = 0; i < n; ++i)
            ptr[i] = v.ptr[i];
        std::cout << __PRETTY_FUNCTION__ << '\n';
    }


//  Move Copy-Constructor
	Vector(Vector&& v) :
        n {v.n}, ptr {v.ptr}
    {
		// lasam ob v intr-o stare consistenta
		v.n = 0;
		v.ptr = nullptr;
		
		std::cout << __PRETTY_FUNCTION__ << '\n';
    }
   
    // Assignment Operator (atribuire)
    Vector& operator = (const Vector& v)
    {
        if (&v == this)    // pt cazul autoatribuirii
            return *this;

        delete[] ptr;      // Renuntarea la resurse

        n = v.n;
        ptr = new T[n];    // Achizitionam noi resurse

        for (size_t i = 0; i < n; ++i)
            ptr[i] = v.ptr[i];
		std::cout << __PRETTY_FUNCTION__ << '\n';
        return *this;
    }

	// Move Assignment Operator (atribuire)
	// Atribuire - Mutare
    Vector& operator = (Vector&& v)
    {
        if (&v == this)    // pt cazul autoatribuirii
            return *this;

        delete[] ptr;      // Renuntarea la resurse

        n = v.n;
        ptr = v.ptr;       // Mutam resursele
        
        // Obiectul v trebuie lasat intr-o stare CONSISTENTA
        v.n = 0;
        v.ptr = nullptr;
        std::cout << __PRETTY_FUNCTION__ << '\n';
        
        return *this;
    }

    // Destructorul elibereaza resursele
   ~Vector()
    {
       delete[] ptr;
    }
//========== RAII================================/

    // Functii membre

    iterator begin() const
    {
        return ptr;
    }

    iterator end() const
    {
        return ptr + n;
    }


private:
    size_t n;
    T* ptr;
};

template <class T>
void Print(const Vector<T>& v)
{
    for (auto& x : v)
        std::cout << x << ' ';
    std::cout << '\n';
}

int main()
{
    Vector<int> v1(10, -7);
    
    {
	//	Vector<int> v2 = std::move(v1); // copiere din v1
	//	Vector<int> v2 = v1;
	}
	
	{
		Vector<int> v2(20, -8);
	//	v1 = v2;             // atribuire   
	//	v1 = std::move(v2);
		v1 = Vector<int>(1000, -2);
	}

//	Print(v1);

    return 0;
}
